# -*- coding: utf-8 -*-
from __future__ import annotations

from typing import Collection, TYPE_CHECKING

from qtpy.QtCore import QModelIndex, Qt, Signal, Slot
from qtpy.QtWidgets import QDialog, QDialogButtonBox, QFormLayout, QListWidget, QListWidgetItem, QVBoxLayout, QWidget

from .html_style_delegate import HTMLDelegate
from .selectable_label import SelectableLabel
from .url_label import URLLabel
from ..catalog import Catalog
from ..utils import HUMAN_READABLE, ID, INCHI_KEY, LINES, SPECIES_TAG, STATE_HTML, best_name, chem_html

__all__ = ['SubstanceInfoSelector', 'SubstanceInfo']


class SubstanceInfoSelector(QDialog):
    tagSelectionChanged: Signal = Signal(int, bool, name='tagSelectionChanged')

    def __init__(self, catalog: Catalog, species_tags: Collection[int], *,
                 selected_species_tags: Collection[int] = (),
                 allow_html: bool = True, inchi_key_search_url_template: str = '',
                 parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self._catalog: Catalog = catalog
        self._inchi_key_search_url_template: str = inchi_key_search_url_template
        self.setModal(True)
        self.setWindowTitle(self.tr('Select Substance'))
        if parent is not None:
            self.setWindowIcon(parent.windowIcon())
        layout: QVBoxLayout = QVBoxLayout(self)
        self._list_box: QListWidget = QListWidget(self)
        self._list_box.itemChanged.connect(self._on_list_item_changed)
        self._list_box.doubleClicked.connect(self._on_list_double_clicked)
        self._list_box.setItemDelegateForColumn(0, HTMLDelegate())
        layout.addWidget(self._list_box)
        species_tags = set(species_tags)
        for entry in catalog.catalog:
            if not species_tags:
                break  # nothing to search for
            if (species_tag := entry[SPECIES_TAG]) in species_tags:
                species_tags.discard(species_tag)  # don't search for the SPECIES_TAG again
                # don't specify the parent here: https://t.me/qtforpython/20950
                item: QListWidgetItem = QListWidgetItem(best_name(entry, allow_html=allow_html))
                item.setData(Qt.ItemDataRole.ToolTipRole, str(species_tag))
                item.setData(Qt.ItemDataRole.UserRole, species_tag)
                item.setCheckState(Qt.CheckState.Checked
                                   if species_tag in selected_species_tags
                                   else Qt.CheckState.Unchecked)
                self._list_box.addItem(item)
        self._buttons: QDialogButtonBox = QDialogButtonBox(QDialogButtonBox.StandardButton.Close, self)
        self._buttons.rejected.connect(self.reject)
        layout.addWidget(self._buttons)

    @Slot(QListWidgetItem)
    def _on_list_item_changed(self, item: QListWidgetItem) -> None:
        self.tagSelectionChanged.emit(item.data(Qt.ItemDataRole.UserRole), item.checkState() == Qt.CheckState.Checked)

    @Slot(QModelIndex)
    def _on_list_double_clicked(self, index: QModelIndex) -> None:
        item: QListWidgetItem = self._list_box.item(index.row())
        syn: SubstanceInfo = SubstanceInfo(self._catalog, item.data(Qt.ItemDataRole.UserRole),
                                           inchi_key_search_url_template=self._inchi_key_search_url_template,
                                           parent=self)
        syn.exec()


class SubstanceInfo(QDialog):
    """ A simple dialog that displays the information about a substance from the loaded catalog """

    def __init__(self, catalog: Catalog, species_tag: int, inchi_key_search_url_template: str = '',
                 parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.setModal(True)
        self.setWindowTitle(self.tr('Substance Info'))
        if parent is not None:
            self.setWindowIcon(parent.windowIcon())
        layout: QFormLayout = QFormLayout(self)
        label: SelectableLabel
        for entry in catalog.catalog:
            if entry[SPECIES_TAG] == species_tag:
                for key in entry:
                    if key == LINES:
                        continue
                    elif key == ID:
                        label = URLLabel(
                            url=f'https://cdms.astro.uni-koeln.de/cdms/portal/catalog/{entry[key]}/',
                            text=f'{entry[key]}',
                            parent=self)
                        label.setOpenExternalLinks(True)
                    elif key == STATE_HTML:
                        label = SelectableLabel(chem_html(str(entry[key])), self)
                    elif key == INCHI_KEY and inchi_key_search_url_template:
                        label = URLLabel(
                            url=inchi_key_search_url_template.format(InChIKey=entry[key]),
                            text=entry[key],
                            parent=self)
                        label.setOpenExternalLinks(True)
                    else:
                        label = SelectableLabel(str(entry[key]), self)
                    layout.addRow(self.tr(HUMAN_READABLE[key]), label)
                label = SelectableLabel(str(len(entry[LINES])), self)
                layout.addRow(self.tr('Number of spectral lines'), label)
                break
        buttons: QDialogButtonBox = QDialogButtonBox(QDialogButtonBox.StandardButton.Close, self)
        buttons.rejected.connect(self.reject)
        layout.addWidget(buttons)

        # add the texts to the translation table but don't run the code at runtime
        if TYPE_CHECKING:
            self.tr('Catalog')
            self.tr('Lines')
            self.tr('Frequency')
            self.tr('Intensity')
            self.tr('ID')
            self.tr('Molecule')
            self.tr('Structural formula')
            self.tr('Stoichiometric formula')
            self.tr('Molecule symbol')
            self.tr('Species tag')
            self.tr('Name')
            self.tr('Trivial name')
            self.tr('Isotopolog')
            self.tr('State (TeX)')
            self.tr('State (HTML)')
            self.tr('InChI key')
            self.tr('Contributor')
            self.tr('Version')
            self.tr('Date of entry')
            self.tr('Degrees of freedom')
            self.tr('Lower state energy')



# -*- coding: utf-8 -*-
from __future__ import annotations

import abc
from pathlib import Path

from qtpy.QtWidgets import QDialog, QMessageBox, QWidget, QWizard
from qtpy.compat import getsavefilename

from ..waiting_screen import WaitingScreen
from ...catalog import CatalogEntryType
from ...utils import ensure_prefix, save_catalog_to_file

__all__ = ['SaveCatalogWizard']


class SaveCatalogWaitingScreen(WaitingScreen):
    def __init__(self, parent: QWidget, *, filename, catalog, frequency_limits, margins: float | None = None) -> None:
        super(QWidget, self).__init__(parent)
        super().__init__(parent,
                         label=self.tr('Please wait…'),
                         target=save_catalog_to_file,
                         kwargs=dict(filename=filename, catalog=catalog, frequency_limits=frequency_limits),
                         margins=margins)


class SaveCatalogWizard(QWizard):
    def __init__(self, default_save_location: Path | None = None,
                 parent: QWidget | None = None) -> None:
        super().__init__(parent)

        self.catalog: list[CatalogEntryType] = []
        self.default_save_location: Path | None = default_save_location

        self.setModal(True)
        if parent is not None:
            self.setWindowIcon(parent.windowIcon())

    def _get_save_file_name(self, directory: str = '') -> tuple[str, str]:

        formats: dict[tuple[str, ...], str] = {
            ('.json.gz',): self.tr('JSON with GZip compression', 'file type'),
            ('.json.bz2',): self.tr('JSON with Bzip2 compression', 'file type'),
            ('.json.xz', '.json.lzma',): self.tr('JSON with LZMA2 compression', 'file type'),
            ('.json',): self.tr('JSON', 'file type'),
        }

        def join_file_dialog_formats(_formats: dict[tuple[str, ...], str]) -> str:
            f: tuple[str, ...]
            all_supported_extensions: list[str] = []
            for f in _formats.keys():
                all_supported_extensions.extend(ensure_prefix(_f, '*') for _f in f)
            format_lines: list[str] = [''.join((
                self.tr('All supported', 'file type'),
                '(',
                ' '.join(ensure_prefix(_f, '*') for _f in all_supported_extensions),
                ')'))]
            n: str
            for f, n in _formats.items():
                format_lines.append(''.join((n, '(', ' '.join(ensure_prefix(_f, '*') for _f in f), ')')))
            format_lines.append(self.tr('All files', 'file type') + '(* *.*)')
            return ';;'.join(format_lines)

        filename: str
        _filter: str
        filename, _filter = getsavefilename(self,
                                            caption=self.tr('Save As…'),
                                            filters=join_file_dialog_formats(formats),
                                            basedir=directory)
        return filename, _filter

    @abc.abstractmethod
    def frequency_limits(self) -> tuple[float, float]:
        ...

    def done(self, exit_code: QDialog.DialogCode) -> None:
        ws: SaveCatalogWaitingScreen
        if exit_code == QDialog.DialogCode.Accepted and self.catalog:
            if self.default_save_location is not None:
                try:
                    ws = SaveCatalogWaitingScreen(
                        self,
                        filename=self.default_save_location,
                        catalog=self.catalog,
                        frequency_limits=self.frequency_limits()
                    )
                    ws.exec()
                except OSError as ex:
                    QMessageBox.warning(
                        self,
                        self.tr('Unable to save the catalog'),
                        self.tr('Error {exception} occurred while saving {filename}. Try another location.')
                        .format(exception=ex, filename=self.default_save_location)
                    )
                else:
                    return super(SaveCatalogWizard, self).done(exit_code)

            save_file_name: str
            while True:
                save_file_name, _ = self._get_save_file_name()
                if not save_file_name:
                    return super(SaveCatalogWizard, self).done(QDialog.DialogCode.Rejected)

                try:
                    ws = SaveCatalogWaitingScreen(
                        self,
                        filename=save_file_name,
                        catalog=self.catalog,
                        frequency_limits=self.frequency_limits()
                    )
                    ws.exec()
                except OSError as ex:
                    QMessageBox.warning(
                        self,
                        self.tr('Unable to save the catalog'),
                        self.tr('Error {exception} occurred while saving {filename}. Try another location.')
                        .format(exception=ex, filename=save_file_name)
                    )
                else:
                    return super(SaveCatalogWizard, self).done(exit_code)

        return super(SaveCatalogWizard, self).done(exit_code)



# -*- coding: utf-8 -*-
from __future__ import annotations

from qtpy.QtCore import QLocale
from qtpy.QtWidgets import QLabel, QVBoxLayout, QWidget, QWizard, QWizardPage

__all__ = ['DownloadConfirmationPage']


class DownloadConfirmationPage(QWizardPage):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)

        self.setTitle(self.tr('Downloading catalog'))
        self.setCommitPage(True)

        layout: QVBoxLayout = QVBoxLayout(self)
        self._label: QLabel = QLabel(self)
        layout.addWidget(self._label)

    def initializePage(self) -> None:
        super(DownloadConfirmationPage, self).initializePage()
        self.setButtonText(QWizard.WizardButton.CommitButton, self.tr('&Start'))
        self._label.setText(
            self.tr('Click {button_text} to start the download data for {min_frequency} to {max_frequency} MHz.')
            .format(
                button_text=self.buttonText(QWizard.WizardButton.CommitButton).replace('&', ''),
                min_frequency=QLocale().toString(self.field('min_frequency')),
                max_frequency=QLocale().toString(self.field('max_frequency'))
            )
        )



# -*- coding: utf-8 -*-
from __future__ import annotations

from math import inf

from qtpy.QtWidgets import QDialog, QWidget

from .catalog_wizard import SaveCatalogWizard
from .catalog_wizard.download_confirmation_page import DownloadConfirmationPage
from .catalog_wizard.progress_page import ProgressPage
from .catalog_wizard.settings_page import SettingsPage
from .catalog_wizard.summary_page import SummaryPage

__all__ = ['DownloadDialog']


class DownloadDialog(SaveCatalogWizard):
    """ GUI for `async_downloader.Downloader` or `downloader.Downloader` """

    def __init__(self, frequency_limits: tuple[float, float] = (-inf, inf),
                 parent: QWidget | None = None) -> None:
        super().__init__(parent=parent)

        self.setWindowTitle(self.tr('Download Catalog'))

        self._settings_page: SettingsPage = SettingsPage(self)
        self.addPage(self._settings_page)
        self.addPage(DownloadConfirmationPage(self))
        self._progress_page: ProgressPage = ProgressPage(self)
        self.addPage(self._progress_page)
        self.addPage(SummaryPage(self))

        self._settings_page.frequency_limits = frequency_limits

    def back(self) -> None:
        if self._progress_page.isActive():
            self._progress_page.stop()
        super().back()

    def next(self) -> None:
        if self._progress_page.isActive():
            self._progress_page.stop()
        super().next()

    def restart(self) -> None:
        if self._progress_page.isActive():
            self._progress_page.stop()
        super().restart()

    def frequency_limits(self) -> tuple[float, float]:
        return self._settings_page.frequency_limits

    def done(self, exit_code: QDialog.DialogCode) -> None:
        if self._progress_page.isActive():
            self._progress_page.stop()
        return super().done(exit_code)



# -*- coding: utf-8 -*-
from __future__ import annotations

import html
import html.entities
import itertools
import os
import sys
from math import e as _e, inf, log10, nan, pow
from numbers import Real
from typing import Any, Callable, Final, Iterable, Protocol, Sequence, TypeVar, cast, overload

__all__ = ['M_LOG10E',
           'T0', 'c', 'h', 'k', 'e',
           'CATALOG', 'BUILD_TIME', 'LINES', 'FREQUENCY', 'INTENSITY', 'ID', 'STRUCTURAL_FORMULA',
           'STOICHIOMETRIC_FORMULA', 'MOLECULE', 'MOLECULE_SYMBOL', 'SPECIES_TAG', 'NAME', 'TRIVIAL_NAME', 'ISOTOPOLOG',
           'STATE', 'STATE_HTML', 'INCHI_KEY', 'DEGREES_OF_FREEDOM', 'LOWER_STATE_ENERGY', 'CONTRIBUTOR', 'VERSION',
           'DATE_OF_ENTRY', 'HUMAN_READABLE',
           'ghz_to_mhz', 'ghz_to_nm', 'ghz_to_rec_cm',
           'mhz_to_ghz', 'mhz_to_nm', 'mhz_to_rec_cm',
           'nm_to_ghz', 'nm_to_mhz', 'nm_to_rec_cm',
           'rec_cm_to_ghz', 'rec_cm_to_mhz', 'rec_cm_to_nm', 'rec_cm_to_meV', 'rec_cm_to_j',
           'meV_to_rec_cm', 'j_to_rec_cm',
           'log10_sq_nm_mhz_to_sq_nm_mhz',
           'log10_sq_nm_mhz_to_log10_cm_per_molecule',
           'log10_sq_nm_mhz_to_cm_per_molecule',
           'sq_nm_mhz_to_log10_sq_nm_mhz',
           'log10_cm_per_molecule_to_log10_sq_nm_mhz',
           'cm_per_molecule_to_log10_sq_nm_mhz',
           'sort_unique', 'merge_sorted', 'search_sorted',
           'within', 'chem_html', 'best_name', 'remove_html', 'wrap_in_html',
           'ensure_prefix',
           'save_catalog_to_file',
           'ReleaseInfo', 'latest_release', 'update_with_pip']

M_LOG10E: Final[float] = log10(_e)

T0: Final[float] = 300.00  # [K], see https://spec.jpl.nasa.gov/ftp/pub/catalog/doc/catdoc.pdf
k: Final[float] = 1.380649000e-23  # [J/K],  see https://physics.nist.gov/cgi-bin/cuu/Value?k
h: Final[float] = 6.626070150e-34  # [J/Hz], see https://physics.nist.gov/cgi-bin/cuu/Value?h
e: Final[float] = 1.602176634e-19  # [C],    see https://physics.nist.gov/cgi-bin/cuu/Value?e
c: Final[float] = 299_792_458.000  # [m/s],  see https://physics.nist.gov/cgi-bin/cuu/Value?c

CATALOG: Final[str] = 'catalog'
BUILD_TIME: Final[str] = 'build_time'
LINES: Final[str] = 'lines'
FREQUENCY: Final[str] = 'frequency'
INTENSITY: Final[str] = 'intensity'
ID: Final[str] = 'id'
MOLECULE: Final[str] = 'molecule'
STRUCTURAL_FORMULA: Final[str] = 'structuralformula'
STOICHIOMETRIC_FORMULA: Final[str] = 'stoichiometricformula'
MOLECULE_SYMBOL: Final[str] = 'moleculesymbol'
SPECIES_TAG: Final[str] = 'speciestag'
NAME: Final[str] = 'name'
TRIVIAL_NAME: Final[str] = 'trivialname'
ISOTOPOLOG: Final[str] = 'isotopolog'
STATE: Final[str] = 'state'
STATE_HTML: Final[str] = 'state_html'
INCHI_KEY: Final[str] = 'inchikey'
CONTRIBUTOR: Final[str] = 'contributor'
VERSION: Final[str] = 'version'
DATE_OF_ENTRY: Final[str] = 'dateofentry'
DEGREES_OF_FREEDOM: Final[str] = 'degreesoffreedom'
LOWER_STATE_ENERGY: Final[str] = 'lowerstateenergy'

HUMAN_READABLE: Final[dict[str, str]] = {
    CATALOG: 'Catalog',
    LINES: 'Lines',
    FREQUENCY: 'Frequency',
    INTENSITY: 'Intensity',
    ID: 'ID',
    MOLECULE: 'Molecule',
    STRUCTURAL_FORMULA: 'Structural formula',
    STOICHIOMETRIC_FORMULA: 'Stoichiometric formula',
    MOLECULE_SYMBOL: 'Molecule symbol',
    SPECIES_TAG: 'Species tag',
    NAME: 'Name',
    TRIVIAL_NAME: 'Trivial name',
    ISOTOPOLOG: 'Isotopolog',
    STATE: 'State (TeX)',
    STATE_HTML: 'State (HTML)',
    INCHI_KEY: 'InChI key',
    CONTRIBUTOR: 'Contributor',
    VERSION: 'Version',
    DATE_OF_ENTRY: 'Date of entry',
    DEGREES_OF_FREEDOM: 'Degrees of freedom',
    LOWER_STATE_ENERGY: 'Lower state energy',
}


def within(x: float, limits: tuple[float, float] | tuple[tuple[float, float], ...]) -> bool:
    if len(limits) < 2:
        raise ValueError('Invalid limits')
    if all(isinstance(limit, Real) for limit in limits):
        return min(limits) <= x <= max(limits)
    elif all(isinstance(limit, tuple) for limit in limits):
        return any(min(limit) <= x <= max(limit) for limit in limits)
    else:
        raise TypeError('Invalid limits type')


_AnyType = TypeVar('_AnyType')


class SupportsLessAndEqual(Protocol[_AnyType]):
    def __eq__(self, other: _AnyType) -> bool:
        pass

    def __lt__(self, other: _AnyType) -> bool:
        pass

    def __le__(self, other: _AnyType) -> bool:
        pass


@overload
def sort_unique(items: Sequence[_AnyType], *,
                key: Callable[[_AnyType], SupportsLessAndEqual],
                reverse: bool = False) -> list[_AnyType]:
    pass


@overload
def sort_unique(items: Sequence[SupportsLessAndEqual], *,
                key: Callable[[SupportsLessAndEqual], SupportsLessAndEqual] | None = None,
                reverse: bool = False) -> list[SupportsLessAndEqual]:
    pass


def sort_unique(items: Sequence[SupportsLessAndEqual] | Sequence[_AnyType], *,
                key: Callable[[_AnyType], SupportsLessAndEqual] | None = None,
                reverse: bool = False) -> list[SupportsLessAndEqual]:
    sorted_items: list[SupportsLessAndEqual] = sorted(items, key=key, reverse=reverse)
    i: int = 0
    while i < len(sorted_items) - 1:
        while i < len(sorted_items) - 1 and sorted_items[i] == sorted_items[i + 1]:
            del sorted_items[i + 1]
        i += 1
    return sorted_items


@overload
def merge_sorted(items_1: Sequence[_AnyType],
                 items_2: Sequence[_AnyType], *,
                 key: Callable[[_AnyType], SupportsLessAndEqual]) -> list[_AnyType]:
    pass


@overload
def merge_sorted(items_1: Sequence[SupportsLessAndEqual],
                 items_2: Sequence[SupportsLessAndEqual], *,
                 key: Callable[[_AnyType], SupportsLessAndEqual] | None = None) -> list[SupportsLessAndEqual]:
    pass


def merge_sorted(items_1: Sequence[SupportsLessAndEqual] | Sequence[_AnyType],
                 items_2: Sequence[SupportsLessAndEqual] | Sequence[_AnyType], *,
                 key: Callable[[_AnyType], SupportsLessAndEqual] | None = None) -> list[SupportsLessAndEqual]:
    sorted_items_1: list[SupportsLessAndEqual] = sort_unique(items_1, key=key, reverse=False)
    sorted_items_2: list[SupportsLessAndEqual] = sort_unique(items_2, key=key, reverse=False)
    merged_items: list[SupportsLessAndEqual] = []

    last_i_1: int
    last_i_2: int
    i_1: int = 0
    i_2: int = 0

    if key is None:
        def key(value: _AnyType) -> SupportsLessAndEqual:
            return value

    while i_1 < len(sorted_items_1) and i_2 < len(sorted_items_2):
        last_i_1 = i_1
        while (i_1 < len(sorted_items_1) and i_2 < len(sorted_items_2)
               and key(sorted_items_1[i_1]) <= key(sorted_items_2[i_2])):
            i_1 += 1
        if last_i_1 < i_1:
            merged_items.extend(sorted_items_1[last_i_1:i_1])

        last_i_2 = i_2
        while (i_1 < len(sorted_items_1) and i_2 < len(sorted_items_2)
               and key(sorted_items_2[i_2]) <= key(sorted_items_1[i_1])):
            i_2 += 1
        if last_i_2 < i_2:
            merged_items.extend(sorted_items_2[last_i_2:i_2])

    while i_1 < len(sorted_items_1) and key(merged_items[-1]) == key(sorted_items_1[i_1]):
        i_1 += 1
    if i_1 < len(sorted_items_1) and i_2 >= len(sorted_items_2):
        merged_items.extend(sorted_items_1[i_1:])
    while i_2 < len(sorted_items_2) and key(merged_items[-1]) == key(sorted_items_2[i_2]):
        i_2 += 1
    if i_2 < len(sorted_items_2) and i_1 >= len(sorted_items_1):
        merged_items.extend(sorted_items_2[i_2:])
    return merged_items


@overload
def search_sorted(threshold: SupportsLessAndEqual,
                  items: Sequence[_AnyType], *,
                  key: Callable[[_AnyType], SupportsLessAndEqual],
                  maybe_equal: bool = False) -> int:
    pass


@overload
def search_sorted(threshold: SupportsLessAndEqual,
                  items: Sequence[SupportsLessAndEqual], *,
                  key: Callable[[_AnyType], SupportsLessAndEqual] | None = None,
                  maybe_equal: bool = False) -> int:
    pass


def search_sorted(threshold: SupportsLessAndEqual,
                  items: Sequence[_AnyType] | Sequence[SupportsLessAndEqual], *,
                  key: Callable[[_AnyType], SupportsLessAndEqual] | None = None,
                  maybe_equal: bool = False) -> int:
    from operator import lt, le

    if not items:
        raise ValueError('Empty sequence provided')
    if key is None:
        def key(value: _AnyType) -> SupportsLessAndEqual:
            return value

    less: Callable[[SupportsLessAndEqual, SupportsLessAndEqual], bool] = le if maybe_equal else lt
    if not less(key(items[0]), threshold):
        return -1
    if less(key(items[-1]), threshold):
        return len(items)
    i: int = 0
    j: int = len(items) - 1
    n: int
    while j - i > 1:
        n = (i + j) // 2
        if less(key(items[n]), threshold):
            i = n
        else:
            j = n
    if i != j and not less(key(items[j]), threshold):
        return i
    return j


def mhz_to_ghz(frequency_mhz: float) -> float:
    return frequency_mhz * 1e-3


def mhz_to_rec_cm(frequency_mhz: float) -> float:
    return frequency_mhz * 1e4 / c


def mhz_to_nm(frequency_mhz: float) -> float:
    return c / frequency_mhz * 1e3


def ghz_to_mhz(frequency_ghz: float) -> float:
    return frequency_ghz * 1e3


def ghz_to_rec_cm(frequency_ghz: float) -> float:
    return frequency_ghz * 1e7 / c


def ghz_to_nm(frequency_ghz: float) -> float:
    return c / frequency_ghz


def rec_cm_to_mhz(frequency_rec_cm: float) -> float:
    return frequency_rec_cm * 1e-4 * c


def rec_cm_to_ghz(frequency_rec_cm: float) -> float:
    return frequency_rec_cm * 1e-7 * c


def rec_cm_to_nm(frequency_rec_cm: float) -> float:
    return 1e7 / frequency_rec_cm


def rec_cm_to_meV(energy_rec_cm: float) -> float:
    return 1e5 * h * c / e * energy_rec_cm


def rec_cm_to_j(energy_rec_cm: float) -> float:
    return 1e2 * h * c * energy_rec_cm


def nm_to_mhz(frequency_nm: float) -> float:
    return c / frequency_nm * 1e-3


def nm_to_ghz(frequency_nm: float) -> float:
    return c / frequency_nm


def nm_to_rec_cm(frequency_nm: float) -> float:
    return 1e7 / frequency_nm


def meV_to_rec_cm(energy_mev: float) -> float:
    return 1e-5 * e / h / c * energy_mev


def j_to_rec_cm(energy_j: float) -> float:
    return 1e-2 / h / c * energy_j


def log10_sq_nm_mhz_to_sq_nm_mhz(intensity_log10_sq_nm_mhz: float) -> float:
    return pow(10.0, intensity_log10_sq_nm_mhz)


def log10_sq_nm_mhz_to_log10_cm_per_molecule(intensity_log10_sq_nm_mhz: float) -> float:
    return -10. + intensity_log10_sq_nm_mhz - log10(c)


def log10_sq_nm_mhz_to_cm_per_molecule(intensity_log10_sq_nm_mhz: float) -> float:
    return pow(10.0, log10_sq_nm_mhz_to_log10_cm_per_molecule(intensity_log10_sq_nm_mhz))


def sq_nm_mhz_to_log10_sq_nm_mhz(intensity_sq_nm_mhz: float) -> float:
    if intensity_sq_nm_mhz == 0.0:
        return -inf
    if intensity_sq_nm_mhz < 0.0:
        return nan
    return log10(intensity_sq_nm_mhz)


def log10_cm_per_molecule_to_log10_sq_nm_mhz(intensity_log10_cm_per_molecule: float) -> float:
    return intensity_log10_cm_per_molecule + 10. + log10(c)


def cm_per_molecule_to_log10_sq_nm_mhz(intensity_cm_per_molecule: float) -> float:
    if intensity_cm_per_molecule == 0.0:
        return -inf
    if intensity_cm_per_molecule < 0.0:
        return nan
    return log10_cm_per_molecule_to_log10_sq_nm_mhz(log10(intensity_cm_per_molecule))


def tex_to_html_entity(s: str) -> str:
    r"""
    Change LaTeX entities syntax to HTML one.
    Get ‘\alpha’ to be ‘&alpha;’ and so on. Unknown LaTeX entities do not get replaced.

    :param s: a line to convert
    :return: a line with all LaTeX entities renamed
    """
    word_start: int = -1
    word_started: bool = False
    backslash_found: bool = False
    _i: int = 0
    fixes: dict[str, str] = {
        'neq': '#8800',
    }
    while _i < len(s):
        _c: str = s[_i]
        if word_started and not _c.isalpha():
            word_started = False
            if s[word_start:_i] + ';' in html.entities.entitydefs:
                s = s[:word_start - 1] + '&' + s[word_start:_i] + ';' + s[_i:]
                _i += 2
            elif s[word_start:_i] in fixes:
                s = s[:word_start - 1] + '&' + fixes[s[word_start:_i]] + ';' + s[_i:]
                _i += 2
        if backslash_found and _c.isalpha() and not word_started:
            word_start = _i
            word_started = True
        backslash_found = (_c == '\\')
        _i += 1
    if word_started:
        if s[word_start:_i] + ';' in html.entities.entitydefs:
            s = s[:word_start - 1] + '&' + s[word_start:_i] + ';' + s[_i:]
            _i += 2
        elif s[word_start:_i] in fixes:
            s = s[:word_start - 1] + '&' + fixes[s[word_start:_i]] + ';' + s[_i:]
            _i += 2
    return s


def chem_html(formula: str) -> str:
    """ converts plain text chemical formula into html markup """
    if '<' in formula or '>' in formula:
        # we can not tell whether it's a tag or a mathematical sign
        return formula

    def sub_tag(s: str) -> str:
        return '<sub>' + s + '</sub>'

    def sup_tag(s: str) -> str:
        return '<sup>' + s + '</sup>'

    def i_tag(s: str) -> str:
        return '<i>' + s + '</i>'

    def subscript(s: str) -> str:
        number_start: int = -1
        number_started: bool = False
        cap_alpha_started: bool = False
        low_alpha_started: bool = False
        _i: int = 0
        while _i < len(s):
            _c: str = s[_i]
            if number_started and not _c.isdigit():
                number_started = False
                s = s[:number_start] + sub_tag(s[number_start:_i]) + s[_i:]
                _i += 1
            if (cap_alpha_started or low_alpha_started) and _c.isdigit() and not number_started:
                number_start = _i
                number_started = True
            if low_alpha_started:
                cap_alpha_started = False
                low_alpha_started = False
            if cap_alpha_started and _c.islower() or _c == ')':
                low_alpha_started = True
            cap_alpha_started = _c.isupper()
            _i += 1
        if number_started:
            s = s[:number_start] + sub_tag(s[number_start:])
        return s

    def prefix(s: str) -> str:
        no_digits: bool = False
        _i: int = len(s)
        while not no_digits:
            _i = s.rfind('-', 0, _i)
            if _i == -1:
                break
            if s[:_i].isalpha() and s[:_i].isupper():
                break
            no_digits = True
            _c: str
            unescaped_prefix: str = html.unescape(s[:_i])
            for _c in unescaped_prefix:
                if _c.isdigit() or _c == '<':
                    no_digits = False
                    break
            if no_digits and (unescaped_prefix[0].islower() or unescaped_prefix[0] == '('):
                return i_tag(s[:_i]) + s[_i:]
        return s

    def charge(s: str) -> str:
        if s[-1] in '+-':
            return s[:-1] + sup_tag(s[-1])
        return s

    def v(s: str) -> str:
        if '=' not in s:
            return s[0] + ' = ' + s[1:]
        ss: list[str] = list(map(str.strip, s.split('=')))
        for _i in range(len(ss)):
            if ss[_i].startswith('v'):
                ss[_i] = ss[_i][0] + sub_tag(ss[_i][1:])
        return ' = '.join(ss)

    html_formula: str = html.escape(formula)
    html_formula_pieces: list[str] = list(map(str.strip, html_formula.split(',')))
    for i in range(len(html_formula_pieces)):
        if html_formula_pieces[i].startswith('v'):
            html_formula_pieces = html_formula_pieces[:i] + [', '.join(html_formula_pieces[i:])]
            break
    for i in range(len(html_formula_pieces)):
        if html_formula_pieces[i].startswith('v'):
            html_formula_pieces[i] = v(html_formula_pieces[i])
            break
        for function in (subscript, prefix, charge):
            html_formula_pieces[i] = function(html_formula_pieces[i])
    html_formula = ', '.join(html_formula_pieces)
    return html_formula


def is_good_html(text: str) -> bool:
    """ Basic check that all tags are sound """
    _1, _2, _3 = text.count('<'), text.count('>'), 2 * text.count('</')
    return _1 == _2 and _1 == _3


def best_name(entry: dict[str, int | str | list[dict[str, float]]],
              allow_html: bool = True) -> str:
    species_tag: int = cast(int, entry.get(SPECIES_TAG, 0))
    last: str = best_name.__dict__.get('last', dict()).get(species_tag, dict()).get(allow_html, '')
    if last:
        return last

    def _best_name() -> str:
        if isotopolog := entry.get(ISOTOPOLOG, ''):
            if allow_html:
                if (is_good_html(str(molecule_symbol := entry[MOLECULE_SYMBOL]))
                        and (entry.get(STRUCTURAL_FORMULA, '') == isotopolog
                             or entry.get(STOICHIOMETRIC_FORMULA, '') == isotopolog)):
                    if state_html := cast(str, entry.get(STATE_HTML, '')):
                        # span tags are needed when the molecule symbol is malformed
                        return f'<span>{molecule_symbol}</span>, ' \
                               f'{chem_html(tex_to_html_entity(str(state_html)))}'
                    return str(molecule_symbol)
                else:
                    if state_html := cast(str, entry.get(STATE_HTML, '')):
                        return f'{chem_html(str(isotopolog))}, ' \
                               f'{chem_html(tex_to_html_entity(str(state_html)))}'
                    return chem_html(str(isotopolog))
            else:
                if state_html := cast(str, entry.get(STATE_HTML, '')):
                    return f'{isotopolog}, {remove_html(tex_to_html_entity(state_html))}'
                if state := cast(str, entry.get(STATE, '')):
                    return f'{isotopolog}, {remove_html(tex_to_html_entity(state.strip("$")))}'
                return isotopolog

        for key in (NAME, STRUCTURAL_FORMULA, STOICHIOMETRIC_FORMULA):
            if candidate := entry.get(key, ''):
                return chem_html(str(candidate)) if allow_html else str(candidate)
        if trivial_name := entry.get(TRIVIAL_NAME, ''):
            return str(trivial_name)
        if species_tag:
            return str(species_tag)
        return 'no name'

    res: str = _best_name()
    if not species_tag:
        return res
    if 'last' not in best_name.__dict__:
        best_name.__dict__['last'] = dict()
    if species_tag not in best_name.__dict__['last']:
        best_name.__dict__['last'][species_tag] = dict()
    best_name.__dict__['last'][species_tag][allow_html] = res
    return res


def remove_html(line: str) -> str:
    """ removes HTML tags and decodes HTML entities """
    if not is_good_html(line):
        return html.unescape(line)

    new_line: str = line
    tag_start: int = new_line.find('<')
    tag_end: int = new_line.find('>', tag_start)
    while tag_start != -1 and tag_end != -1:
        new_line = new_line[:tag_start] + new_line[tag_end + 1:]
        tag_start = new_line.find('<')
        tag_end = new_line.find('>', tag_start)
    return html.unescape(new_line).lstrip()


def wrap_in_html(text: str, line_end: str = os.linesep) -> str:
    """ Make a full HTML document out of a piece of the markup """
    new_text: list[str] = [
        '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0 Transitional//EN">',
        '<html lang="en" xml:lang="en">',
        '<head>',
        '<meta http-equiv="content-type" content="text/html; charset=utf-8">',
        '</head>',
        '<body>',
        text,
        '</body>',
        '</html>'
    ]

    return line_end.join(new_text)


def ensure_prefix(text: str, prefix: str) -> str:
    if text.startswith(prefix):
        return text
    else:
        return prefix + text


def save_catalog_to_file(filename: str | os.PathLike[str],
                         catalog: list[dict[str, int | str | list[dict[str, float]]]],
                         frequency_limits: tuple[float, float]) -> bool:
    from .catalog import Catalog

    if not catalog:
        return False
    Catalog.save(filename=filename, catalog=catalog, frequency_limits=frequency_limits)
    return True


class ReleaseInfo:
    def __init__(self, version: str = '', pub_date: str = '') -> None:
        self.version: str = version
        self.pub_date: str = pub_date

    def __bool__(self) -> bool:
        return bool(self.version) and bool(self.pub_date)

    def __lt__(self, other: str | ReleaseInfo) -> bool:
        if isinstance(other, str):
            other = ReleaseInfo(version=other)
        i: str
        j: str
        for i, j in itertools.zip_longest(self.version.replace('-', '.').split('.'),
                                          other.version.replace('-', '.').split('.'), fillvalue=''):
            if i == j:
                continue
            if i.isdigit() and j.isdigit():
                return int(i) < int(j)
            else:
                i_digits: str = ''.join(itertools.takewhile(str.isdigit, i))
                j_digits: str = ''.join(itertools.takewhile(str.isdigit, j))
                if i_digits != j_digits:
                    if i_digits and j_digits:
                        return int(i_digits) < int(j_digits)
                    else:
                        return i_digits < j_digits
                return i < j
        return False

    def __eq__(self, other: str | ReleaseInfo) -> bool:
        if not isinstance(other, (str, ReleaseInfo)):
            return NotImplemented
        if isinstance(other, str):
            other = ReleaseInfo(version=other)
        return self.version == other.version


def latest_release() -> ReleaseInfo:
    import urllib.request
    import xml.dom.minidom as dom
    from http.client import HTTPResponse
    from xml.dom.minicompat import NodeList
    from urllib.error import URLError

    from . import __original_name__

    try:
        r: HTTPResponse = urllib.request.urlopen(f'https://pypi.org/rss/project/{__original_name__}/releases.xml',
                                                 timeout=1)
    except URLError:
        return ReleaseInfo()
    if r.getcode() != 200 or not r.readable():
        return ReleaseInfo()
    rss: dom.Node | None = dom.parseString(r.read().decode(encoding='ascii')).documentElement
    if not isinstance(rss, dom.Element) or rss.tagName != 'rss':
        return ReleaseInfo()
    channels: NodeList = rss.getElementsByTagName('channel')
    if not channels or channels[0].nodeType != dom.Node.ELEMENT_NODE:
        return ReleaseInfo()
    channel: dom.Element = channels[0]
    items: NodeList = channel.getElementsByTagName('item')
    if not items or items[0].nodeType != dom.Node.ELEMENT_NODE:
        return ReleaseInfo()
    item: dom.Element = items[0]
    titles: NodeList = item.getElementsByTagName('title')
    if not titles or titles[0].nodeType != dom.Node.ELEMENT_NODE:
        return ReleaseInfo()
    title: dom.Element = titles[0]
    pub_dates: NodeList = item.getElementsByTagName('pubDate')
    if not pub_dates or pub_dates[0].nodeType != dom.Node.ELEMENT_NODE:
        return ReleaseInfo()
    pub_date: dom.Element = pub_dates[0]
    title_value: dom.Node = title.firstChild
    pub_date_value: dom.Node = pub_date.firstChild
    if not isinstance(title_value, dom.Text) or not isinstance(pub_date_value, dom.Text):
        return ReleaseInfo()

    return ReleaseInfo(title_value.data, pub_date_value.data)


def update_with_pip() -> None:
    import subprocess
    import sys

    from . import __original_name__

    subprocess.Popen(args=[
        sys.executable, '-c',
        f'''import sys, subprocess, time; time.sleep(2);\
        subprocess.run(args=[sys.executable, '-m', 'pip', 'install', '-U', {__original_name__!r}]);\
        subprocess.Popen(args=[sys.executable, '-m', {__original_name__!r}])'''])
    sys.exit(0)


if sys.version_info < (3, 10):
    import builtins


    # noinspection PyShadowingBuiltins, PyUnusedLocal
    def zip(*iterables: Iterable[Any], strict: bool = False) -> builtins.zip:
        """ Intentionally override `builtins.zip` to ignore `strict` parameter in Python < 3.10 """
        return builtins.zip(*iterables)


    __all__.append('zip')



# -*- coding: utf-8 -*-
from __future__ import annotations

from math import inf

from qtpy.QtWidgets import QDoubleSpinBox, QFormLayout, QWidget, QWizardPage

__all__ = ['SettingsPage']


class SettingsPage(QWizardPage):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)

        self.setTitle(self.tr('New catalog'))

        layout: QFormLayout = QFormLayout(self)

        self._spin_min_frequency: QDoubleSpinBox = QDoubleSpinBox(self)
        self._spin_max_frequency: QDoubleSpinBox = QDoubleSpinBox(self)
        self._spin_min_frequency.setRange(0., inf)
        self._spin_max_frequency.setRange(0., inf)
        self._spin_min_frequency.valueChanged.connect(self._spin_max_frequency.setMinimum)
        self._spin_max_frequency.valueChanged.connect(self._spin_min_frequency.setMaximum)
        self._spin_min_frequency.setPrefix(self.tr('', 'spin prefix'))
        self._spin_max_frequency.setPrefix(self.tr('', 'spin prefix'))
        self._spin_min_frequency.setSuffix(self.tr(' MHz', 'spin suffix'))
        self._spin_max_frequency.setSuffix(self.tr(' MHz', 'spin suffix'))
        layout.addRow(self.tr('Minimal frequency:'), self._spin_min_frequency)
        layout.addRow(self.tr('Maximal frequency:'), self._spin_max_frequency)

        try:  # PyQt*
            self.registerField('min_frequency', self._spin_min_frequency,
                               'value', self._spin_min_frequency.valueChanged)
            self.registerField('max_frequency', self._spin_max_frequency,
                               'value', self._spin_max_frequency.valueChanged)
        except TypeError:  # PySide*
            from qtpy.QtCore import SIGNAL

            self.registerField('min_frequency', self._spin_min_frequency,
                               'value', SIGNAL('self._spin_min_frequency.valueChanged'))
            self.registerField('max_frequency', self._spin_max_frequency,
                               'value', SIGNAL('self._spin_max_frequency.valueChanged'))

    @property
    def frequency_limits(self) -> tuple[float, float]:
        return self._spin_min_frequency.value(), self._spin_max_frequency.value()

    @frequency_limits.setter
    def frequency_limits(self, new_limits: tuple[float, float]) -> None:
        min_frequency: float = min(new_limits)
        max_frequency: float = max(new_limits)
        self._spin_max_frequency.setMaximum(max(2.0 * max_frequency, self._spin_max_frequency.maximum()))
        if min_frequency > self._spin_max_frequency.value():
            self._spin_max_frequency.setValue(max_frequency)
            self._spin_min_frequency.setValue(min_frequency)
        else:
            self._spin_min_frequency.setValue(min_frequency)
            self._spin_max_frequency.setValue(max_frequency)

    def validatePage(self) -> bool:
        return self._spin_max_frequency.value() > self._spin_min_frequency.value()



# -*- coding: utf-8 -*-
from __future__ import annotations

import enum
from pathlib import Path
from typing import Iterable, cast

from qtpy.QtCore import QDateTime, QLocale, QModelIndex, QTimeZone, QUrl, Qt, Signal, Slot
from qtpy.QtGui import QContextMenuEvent, QDesktopServices
from qtpy.QtWidgets import (QAbstractItemView, QAbstractScrollArea, QDialog, QDialogButtonBox, QFormLayout, QMenu,
                            QTableWidget, QTableWidgetItem, QVBoxLayout, QWidget)

from . import qta_icon
from .selectable_label import SelectableLabel
from .titled_list_widget import TitledListWidget
from .update_dialog import UpdateDialog
from ..catalog import Catalog, CatalogSourceInfo

__all__ = ['CatalogInfo']


class SourcesList(QTableWidget):
    catalogUpdated: Signal = Signal(name='catalogUpdated')

    class Columns(enum.IntEnum):
        FileLocation = 0
        BuildTime = 1

    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.setAlternatingRowColors(True)
        self.setColumnCount(len(SourcesList.Columns.__members__))
        self.setCornerButtonEnabled(False)
        self.setDragDropOverwriteMode(False)
        self.setDropIndicatorShown(False)
        self.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)
        self.setHorizontalHeaderLabels([self.tr('Filename'), self.tr('Build Time')])
        self.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.setSortingEnabled(True)
        self.horizontalHeader().setHighlightSections(False)
        self.setHorizontalScrollMode(QAbstractItemView.ScrollMode.ScrollPerPixel)
        self.setSizeAdjustPolicy(QAbstractScrollArea.SizeAdjustPolicy.AdjustToContents)
        self.verticalHeader().setVisible(False)
        self.verticalHeader().setHighlightSections(False)

        self._context_menu: QMenu = QMenu(self)
        self._context_menu.addAction(
            qta_icon('mdi6.target'),
            self.tr('Open File &Location'),
            self._on_open_file_location_triggered
        )
        self._context_menu.setDefaultAction(self._context_menu.actions()[0])
        self._context_menu.addSeparator()
        self._context_menu.addAction(
            qta_icon('mdi6.update'),
            self.tr('&Update Catalog'),
            self._on_update_catalog_triggered
        )

        self.cellDoubleClicked.connect(self._on_item_double_clicked)

    def contextMenuEvent(self, event: QContextMenuEvent) -> None:
        if not self.selectedIndexes():
            return super().contextMenuEvent(event)
        self._context_menu.exec(event.globalPos())
        return super().contextMenuEvent(event)

    def _selected_row(self) -> int | None:
        if not self.selectedIndexes():
            return None
        return cast(QModelIndex, self.selectedIndexes()[0]).row()

    def _file_location(self, row: int) -> Path | None:
        item: QTableWidgetItem | None = self.item(row, SourcesList.Columns.FileLocation)
        if item is None:
            return None
        return item.data(Qt.ItemDataRole.UserRole)

    def _locate_file(self, row: int) -> bool:
        file_location: Path | None = self._file_location(row)
        if file_location is None:
            return False
        return QDesktopServices.openUrl(QUrl.fromLocalFile(str(file_location.parent)))

    @Slot(int, int)
    def _on_item_double_clicked(self, row: int, _column: int) -> None:
        self._locate_file(row)

    @Slot()
    def _on_open_file_location_triggered(self) -> None:
        row: int | None = self._selected_row()
        if row is None:
            return
        self._locate_file(row)

    @Slot()
    def _on_update_catalog_triggered(self) -> None:
        row: int | None = self._selected_row()
        if row is None:
            return
        file_location: Path | None = self._file_location(row)
        if file_location is None:
            return
        ud: UpdateDialog = UpdateDialog(file_location, parent=self.parent())
        if ud.exec():
            self.catalogUpdated.emit()

    def extend(self, info: Iterable[CatalogSourceInfo]) -> None:
        row: int
        info_item: CatalogSourceInfo
        item: QTableWidgetItem
        for row, info_item in enumerate(info, start=self.rowCount()):
            self.setRowCount(row + 1)
            item = QTableWidgetItem(str(info_item.filename))
            item.setToolTip(str(info_item.filename))
            item.setData(Qt.ItemDataRole.UserRole, info_item.filename)
            self.setItem(row, SourcesList.Columns.FileLocation, item)
            if info_item.build_datetime is not None:
                qt_datetime: QDateTime = QDateTime(info_item.build_datetime)
                qt_datetime.setTimeZone(QTimeZone(round(info_item.build_datetime.utcoffset().total_seconds())))
                item = QTableWidgetItem(QLocale().toString(qt_datetime))
                self.setItem(row, SourcesList.Columns.BuildTime, item)
        self.setColumnHidden(SourcesList.Columns.BuildTime,
                             all(self.item(row, SourcesList.Columns.BuildTime) is None
                                 for row in range(self.rowCount()))
                             )
        self.resizeColumnsToContents()


class CatalogInfo(QDialog):
    """ A simple dialog that displays the information about the loaded catalog(s) """

    catalogUpdated: Signal = Signal(name='catalogUpdated')

    def __init__(self, catalog: Catalog, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.setModal(True)
        self.setWindowTitle(self.tr('Catalog Info'))
        if parent is not None:
            self.setWindowIcon(parent.windowIcon())
        layout: QVBoxLayout = QVBoxLayout(self)

        sources_list: SourcesList = SourcesList(self)
        layout.addWidget(sources_list)
        sources_list.extend(catalog.sources_info)
        sources_list.catalogUpdated.connect(self._on_catalog_updated)

        frequency_limits_list: TitledListWidget = TitledListWidget(self)
        frequency_limits_list.setTitle(self.tr('Frequency limits:'))
        layout.addWidget(frequency_limits_list)
        locale: QLocale = QLocale()
        frequency_limits_list.addItems([
            self.tr('{min_frequency} to {max_frequency} MHz')
            .format(min_frequency=locale.toString(min(frequency_limit)),
                    max_frequency=locale.toString(max(frequency_limit)))
            for frequency_limit in catalog.frequency_limits])

        stat_layout: QFormLayout = QFormLayout()
        stat_layout.addRow(self.tr('Total number of substances:'), SelectableLabel(str(catalog.entries_count)))
        layout.addLayout(stat_layout, 0)

        buttons: QDialogButtonBox = QDialogButtonBox(QDialogButtonBox.StandardButton.Close, self)
        buttons.rejected.connect(self.reject)
        layout.addWidget(buttons)

        self.adjustSize()

    @Slot()
    def _on_catalog_updated(self) -> None:
        self.reject()
        self.catalogUpdated.emit()



# -*- coding: utf-8 -*-
from __future__ import annotations

from typing import Callable, Final

from qtpy.QtCore import QAbstractTableModel, QLocale, QModelIndex, QPersistentModelIndex, Qt
from qtpy.QtWidgets import QWidget

from .settings import Settings
from ..catalog import CatalogEntryType
from ..utils import (FREQUENCY, INTENSITY, LINES, LOWER_STATE_ENERGY, SPECIES_TAG, best_name)

__all__ = ['FoundLinesModel']


class FoundLinesModel(QAbstractTableModel):
    ROW_BATCH_COUNT: Final[int] = 5

    class DataType:
        __slots__ = ['species_tag', 'name',
                     'frequency_str', 'frequency',
                     'intensity_str', 'intensity',
                     'lower_state_energy_str', 'lower_state_energy']

        def __init__(self,
                     species_tag: int, name: str,
                     frequency_str: str, frequency: float,
                     intensity_str: str, intensity: float,
                     lower_state_energy_str: str, lower_state_energy: float) -> None:
            self.species_tag: int = species_tag
            self.name: str = name
            self.frequency_str: str = frequency_str
            self.frequency: float = frequency
            self.intensity_str: str = intensity_str
            self.intensity: float = intensity
            self.lower_state_energy_str: str = lower_state_energy_str
            self.lower_state_energy: float = lower_state_energy

        def __eq__(self, other: 'FoundLinesModel.DataType') -> int:
            if not isinstance(other, FoundLinesModel.DataType):
                return NotImplemented
            return (self.species_tag == other.species_tag
                    and self.frequency == other.frequency
                    and self.intensity == other.intensity
                    and self.lower_state_energy == other.lower_state_energy)

        def __hash__(self) -> int:
            return hash(self.species_tag) ^ hash(self.frequency) ^ hash(self.lower_state_energy)

    def __init__(self, settings: Settings, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self._settings: Settings = settings
        self._entries: list[CatalogEntryType] = []
        self._data: list[FoundLinesModel.DataType] = []
        self._rows_loaded: int = FoundLinesModel.ROW_BATCH_COUNT

        unit_format: Final[str] = self.tr('{value} [{unit}]', 'unit format')
        self._header: Final[list[str]] = [
            self.tr('Substance'),
            unit_format.format(value=self.tr('Frequency'), unit=self._settings.frequency_unit_str),
            unit_format.format(value=self.tr('Intensity'), unit=self._settings.intensity_unit_str),
            unit_format.format(value=self.tr('Lower state energy'), unit=self._settings.energy_unit_str),
        ]

    def update_units(self) -> None:
        unit_format: Final[str] = self.tr('{value} [{unit}]', 'unit format')
        self._header[1] = unit_format.format(value=self.tr('Frequency'), unit=self._settings.frequency_unit_str)
        self._header[2] = unit_format.format(value=self.tr('Intensity'), unit=self._settings.intensity_unit_str)
        self._header[3] = unit_format.format(value=self.tr('Lower state energy'), unit=self._settings.energy_unit_str)

    def rowCount(self, parent: QModelIndex | QPersistentModelIndex = ...) -> int:
        return min(len(self._data), self._rows_loaded)

    def columnCount(self, parent: QModelIndex | QPersistentModelIndex = ...) -> int:
        return len(self._header)

    def data(self, index: QModelIndex | QPersistentModelIndex, role: int = Qt.ItemDataRole.DisplayRole) -> str | None:
        if index.isValid():
            if role == Qt.ItemDataRole.DisplayRole:
                item: FoundLinesModel.DataType = self._data[index.row()]
                column_index: int = index.column()
                if column_index == 0:
                    return item.name
                if column_index == 1:
                    return item.frequency_str
                if column_index == 2:
                    return item.intensity_str
                if column_index == 3:
                    return item.lower_state_energy_str
        return None

    def row(self, row_index: int) -> DataType:
        return self._data[row_index]

    def headerData(self, col: int, orientation: Qt.Orientation, role: int = ...) -> str | None:
        if orientation == Qt.Orientation.Horizontal and role == Qt.ItemDataRole.DisplayRole:
            return self._header[col]
        return None

    def setHeaderData(self, section: int, orientation: Qt.Orientation, value: str, role: int = ...) -> bool:
        if (orientation == Qt.Orientation.Horizontal
                and role == Qt.ItemDataRole.DisplayRole
                and 0 <= section < len(self._header)):
            self._header[section] = value
            return True
        return False

    def clear(self) -> None:
        self.set_entries([])

    def set_entries(self, new_data: list[CatalogEntryType]) -> None:
        from_mhz: Callable[[float], float] = self._settings.from_mhz
        from_log10_sq_nm_mhz: Callable[[float], float] = self._settings.from_log10_sq_nm_mhz
        from_rec_cm: Callable[[float], float] = self._settings.from_rec_cm
        frequency_suffix: int = self._settings.frequency_unit
        precision: int = [4, 7, 8, 8][frequency_suffix]
        locale: QLocale = QLocale()
        decimal_point: str = locale.decimalPoint()

        def frequency_str(frequency: float) -> tuple[str, float]:
            frequency = from_mhz(frequency)
            return f'{frequency:.{precision}f}'.replace('.', decimal_point), frequency

        def intensity_str(intensity: float) -> tuple[str, float]:
            intensity = from_log10_sq_nm_mhz(intensity)
            if intensity == 0.0:
                return '0', intensity
            elif abs(intensity) < 0.1:
                return f'{intensity:.4e}'.replace('.', decimal_point), intensity
            else:
                return f'{intensity:.4f}'.replace('.', decimal_point), intensity

        def lower_state_energy_str(lower_state_energy: float) -> tuple[str, float]:
            lower_state_energy = from_rec_cm(lower_state_energy)
            if lower_state_energy == 0.0:
                return '0', lower_state_energy
            elif abs(lower_state_energy) < 0.1:
                return f'{lower_state_energy:.4e}'.replace('.', decimal_point), lower_state_energy
            else:
                return f'{lower_state_energy:.4f}'.replace('.', decimal_point), lower_state_energy

        self.beginResetModel()
        unique_entries: list[CatalogEntryType] = []
        non_unique_indices: set[int] = set()
        unique: bool
        all_unique: bool = True  # unless the opposite is proven
        for i in range(len(new_data)):
            if i in non_unique_indices:
                continue
            unique = True
            for j in range(i + 1, len(new_data)):
                if j in non_unique_indices:
                    continue
                if new_data[i] == new_data[j]:
                    non_unique_indices.add(j)
                    unique = False
                    all_unique = False
                    break
            if unique and not all_unique:
                unique_entries.append(new_data[i])
        if all_unique:
            self._entries = new_data.copy()
        else:
            self._entries = unique_entries
        entry: CatalogEntryType
        rich_text_in_formulas: bool = self._settings.rich_text_in_formulas
        self._data = list(set(
            FoundLinesModel.DataType(
                entry[SPECIES_TAG],
                best_name(entry, rich_text_in_formulas),
                *frequency_str(line[FREQUENCY]),
                *intensity_str(line[INTENSITY]),
                *lower_state_energy_str(line[LOWER_STATE_ENERGY]),
            )
            for entry in self._entries
            for line in entry[LINES]
        ))
        self._rows_loaded = FoundLinesModel.ROW_BATCH_COUNT
        self.endResetModel()

    def sort(self, column: int, order: Qt.SortOrder = Qt.SortOrder.AscendingOrder) -> None:
        self.beginResetModel()
        key = {
            0: (lambda l: (l.name, l.frequency, l.intensity, l.lower_state_energy)),
            1: (lambda l: (l.frequency, l.intensity, l.name, l.lower_state_energy)),
            2: (lambda l: (l.intensity, l.frequency, l.name, l.lower_state_energy)),
            3: (lambda l: (l.lower_state_energy, l.intensity, l.frequency, l.name))
        }[column]
        self._data.sort(key=key, reverse=bool(order != Qt.SortOrder.AscendingOrder))
        self.endResetModel()

    def canFetchMore(self, index: QModelIndex | QPersistentModelIndex = QModelIndex()) -> bool:
        return len(self._data) > self._rows_loaded

    def fetchMore(self, index: QModelIndex | QPersistentModelIndex = QModelIndex()) -> None:
        # https://sateeshkumarb.wordpress.com/2012/04/01/paginated-display-of-table-data-in-pyqt/
        remainder: int = len(self._data) - self._rows_loaded
        items_to_fetch: int = min(remainder, FoundLinesModel.ROW_BATCH_COUNT)
        self.beginInsertRows(QModelIndex(), self._rows_loaded, self._rows_loaded + items_to_fetch - 1)
        self._rows_loaded += items_to_fetch
        self.endInsertRows()
